#include <iostream>
#include <iomanip>
#include <string>
#include <vector>
#include <deque>
#include <memory>
#include <queue>
#include <algorithm>

// How to print our std::vector:
template <typename T>
std::ostream & operator << ( std::ostream & o, std::vector<T> const & v ) {
	auto i = std::begin( v );
	auto end = std::end( v );
	o << '{';
	if ( i != end ) {
		o << ' ' << *i;
		for ( ++i; i != end; ++i ) {
			o << ", " << *i;
		}
		o << ' ';
	}
	o << '}';
	return o;
}

// How to print std::deque:
template <typename T>
std::ostream & operator << ( std::ostream & o, std::deque<T> const & v ) {
	auto i = std::begin( v );
	auto end = std::end( v );
	o << '{';
	if ( i != end ) {
		o << ' ' << *i;
		for ( ++i; i != end; ++i ) {
			o << ", " << *i;
		}
		o << ' ';
	}
	o << '}';
	return o;
}

// Forward declaration for Edge
class Vertex;

// The edges of the graph will be the roads from a simplified
// road map of part of Romania on page 68.
class Edge {
	private:
		double _cost;
		std::shared_ptr<Vertex> _child_vertex_ptr;
	public:
		explicit Edge ( double const & cost, std::shared_ptr<Vertex> const & child_vertex_ptr ) : 
			_cost( cost ),
			_child_vertex_ptr( child_vertex_ptr)
		{
			//
		}
		
		double const & get_cost () const {
			return _cost;
		}
		
		std::shared_ptr<Vertex> const & get_child_vertex_ptr () const {
			return _child_vertex_ptr;
		}
		
		void clear () {
			_cost = 0;
			_child_vertex_ptr = nullptr;

		}
};

// These vertices of a graph will be the cities from the simplified 
// roadmap of part of Romania on page 68.
class Vertex {
	private:
		std::string _name;
		std::vector<std::shared_ptr<Edge>> _edge_ptrs;
		double _g;
		double _h;
		std::shared_ptr<Vertex> _parent;
	public:
		explicit Vertex ( std::string const & name ) :
			_name( name ),
			_edge_ptrs(),
			_g( 0.0 ),
			_h( 0.0 ),
			_parent( nullptr )
		{
			//
		}
		
		std::string const & get_name () const {
			return _name;
		}	
		
		std::vector<std::shared_ptr<Edge>> const & get_edge_ptrs () const {
			return _edge_ptrs;
		}
		
		double const & get_g () const {
			return _g;
		}
		
		double const & get_h () const {
			return _h;
		}
		
		double get_f () const {
			return get_g() + get_h();
		}
		
		std::shared_ptr<Vertex> const & get_parent () const {
			return _parent;
		}
		
		void add_edge_ptr ( double cost, std::shared_ptr<Vertex> const vertex_ptr ) {
			auto edge_ptr = std::make_shared<Edge>( cost, vertex_ptr );
			_edge_ptrs.push_back( edge_ptr );
		}
		
		void set_g ( double value ) {
			_g = value;
		}
		
		void set_h ( double value ) {
			_h = value;
		}
		
		void set_parent ( std::shared_ptr<Vertex> const & value ) {
			_parent = value;
		}
		
		void clear () {
			_name = "";
			auto i = std::begin( _edge_ptrs );
			auto end = std::end( _edge_ptrs );
			while ( i != end ) {
				(*i)->clear();
				++i;
			}
			_g = 0;
			_h = 0;
		}
};

// How to print a std::shared_ptr pointing to an Edge:
std::ostream & operator << ( std::ostream & o, std::shared_ptr<Edge> const & edge_ptr ) {
	o << "{ cost: " << std::setw( 3 ) << edge_ptr->get_cost() << ", vertex_ptr: " << std::setw( 14 ) << edge_ptr->get_child_vertex_ptr()->get_name() << " }";
	return o;
}

// How to print a std::shared_ptr pointing to a Vertex:
std::ostream & operator << ( std::ostream & o, std::shared_ptr<Vertex> const & vertex_ptr ) {
	o << "{ name: " << std::setw( 14 ) << vertex_ptr->get_name() << ", f: " << std::setw( 3 ) << vertex_ptr->get_f() << ", g: " << std::setw( 3 ) << vertex_ptr->get_g() << ", h: " << std::setw( 3 ) << vertex_ptr->get_h();
	//o << ", edge_ptrs: " << vertex_ptr->get_edge_ptrs();
	o << " }";
	return o;
}

// How to print a std::vector of std::shared_ptr pointing to a Vertex:
std::ostream & operator << ( std::ostream & o, std::vector<std::shared_ptr<Vertex>> const & vertex_ptrs ) {
	auto i = std::begin( vertex_ptrs );
	auto end = std::end( vertex_ptrs );
	o << '{';
	if ( i != end ) {
		o << "\n\t" << *i;
		for ( ++i; i != end; ++i ) {
			o << ",\n\t" << *i;
		}
		o << '\n';
	}
	o << '}';
	return o;
}

// How to print a std::deque of std::shared_ptr pointing to a Vertex
std::ostream & operator << ( std::ostream & o, std::deque<std::shared_ptr<Vertex>> const & vertex_ptrs ) {
	auto i = std::begin( vertex_ptrs );
	auto end = std::end( vertex_ptrs );
	o << '{';
	if ( i != end ) {
		o << "\n\t" << *i;
		for ( ++i; i != end; ++i ) {
			o << ",\n\t" << *i;
		}
		o << '\n';
	}
	o << '}';
	return o;
}

void for_each_vertex_set_g_to_0 ( std::vector<std::shared_ptr<Vertex>> & vertex_ptrs ) {
	auto vertex_ptrs_i = std::begin( vertex_ptrs );
	auto vertex_ptrs_end = std::end( vertex_ptrs );
	std::shared_ptr<Vertex> vertex_ptr;
	while ( vertex_ptrs_i != vertex_ptrs_end ) {
		vertex_ptr = *vertex_ptrs_i;
		vertex_ptr->set_g( 0.0 );
		++vertex_ptrs_i;
	}
}

void for_each_vertex_set_h_to_heuristic ( std::vector<std::shared_ptr<Vertex>> & vertex_ptrs ) {
	// Hardcoded to the straight-line distances to Bucharest
	// from Page 93
	// Note: If the coordinates to each of these cities were included
	// in each Vertex, then this could just be a Euclidean distance
	// calculation.
	auto vertex_ptrs_i = std::begin( vertex_ptrs );
	auto vertex_ptrs_end = std::end( vertex_ptrs );
	std::shared_ptr<Vertex> vertex_ptr;
	std::string name;
	while ( vertex_ptrs_i != vertex_ptrs_end ) {
		vertex_ptr = *vertex_ptrs_i;
		name = vertex_ptr->get_name();
		if ( name == "Arad" ) {
			vertex_ptr->set_h( 366 );
		} else if ( name == "Bucharest" ) {
			vertex_ptr->set_h( 0 );
		} else if ( name == "Craiova" ) {
			vertex_ptr->set_h( 160 );
		} else if ( name == "Drobeta" ) {
			vertex_ptr->set_h( 242 );
		} else if ( name == "Eforie" ) {
			vertex_ptr->set_h( 161 );
		} else if ( name == "Fagaras" ) {
			vertex_ptr->set_h( 176 );
		} else if ( name == "Giurgiu" ) {
			vertex_ptr->set_h( 77 );
		} else if ( name == "Hirsova" ) {
			vertex_ptr->set_h( 151 );
		} else if ( name == "Iasi" ) {
			vertex_ptr->set_h( 226 );
		} else if ( name == "Lugoj" ) {
			vertex_ptr->set_h( 244 );
		} else if ( name == "Mehadia" ) {
			vertex_ptr->set_h( 241 );
		} else if ( name == "Neamt" ) {
			vertex_ptr->set_h( 234 );
		} else if ( name == "Oradea" ) {
			vertex_ptr->set_h( 380 );
		} else if ( name == "Pitesti" ) {
			vertex_ptr->set_h( 100 );
		} else if ( name == "Rimnicu Vilcea" ) {
			vertex_ptr->set_h( 193 );
		} else if ( name == "Sibiu" ) {
			vertex_ptr->set_h( 253 );
		} else if ( name == "Timisoara" ) {
			vertex_ptr->set_h( 329 );
		} else if ( name == "Urziceni" ) {
			vertex_ptr->set_h( 80 );
		} else if ( name == "Vaslui" ) {
			vertex_ptr->set_h( 199 );
		} else if ( name == "Zerind" ) {
			vertex_ptr->set_h( 374 );
		} else {
			std::cerr << "ERROR: No match for  name in set_heuristic function\n";
			std::exit( 0 );
		}
		++vertex_ptrs_i;
	}

}

void clear ( std::vector<std::shared_ptr<Vertex>> & graph ) {
	// Clear the vertices and edges of the graph from memory
	auto graph_i = std::begin( graph );
	auto graph_end = std::end( graph );
	std::shared_ptr<Vertex> vertex_ptr;
	while ( graph_i != graph_end ) {
		vertex_ptr = *graph_i;
		vertex_ptr->clear();
		++graph_i;
	}
}

int main ( int argc, char * argv[] ) {
	
	// For each city in the simplified road map of part of Romania on page 68,
	// make a std::shared_ptr to Vertex.
	auto Arad           = std::make_shared<Vertex>( "Arad"           );
	auto Bucharest      = std::make_shared<Vertex>( "Bucharest"      );
	auto Craiova        = std::make_shared<Vertex>( "Craiova"        );
	auto Drobeta        = std::make_shared<Vertex>( "Drobeta"        );
	auto Eforie         = std::make_shared<Vertex>( "Eforie"         );
	auto Fagaras        = std::make_shared<Vertex>( "Fagaras"        );
	auto Giurgiu        = std::make_shared<Vertex>( "Giurgiu"        );
	auto Hirsova        = std::make_shared<Vertex>( "Hirsova"        );
	auto Iasi           = std::make_shared<Vertex>( "Iasi"           );
	auto Lugoj          = std::make_shared<Vertex>( "Lugoj"          );
	auto Mehadia        = std::make_shared<Vertex>( "Mehadia"        );
	auto Neamt          = std::make_shared<Vertex>( "Neamt"          );
	auto Oradea         = std::make_shared<Vertex>( "Oradea"         );
	auto Pitesti        = std::make_shared<Vertex>( "Pitesti"        );
	auto Rimnicu_Vilcea = std::make_shared<Vertex>( "Rimnicu Vilcea" );
	auto Sibiu          = std::make_shared<Vertex>( "Sibiu"          );
	auto Timisoara      = std::make_shared<Vertex>( "Timisoara"      );
	auto Urziceni       = std::make_shared<Vertex>( "Urziceni"       );
	auto Vaslui         = std::make_shared<Vertex>( "Vaslui"         );
	auto Zerind         = std::make_shared<Vertex>( "Zerind"         );

	// Setting the costs from the simplified road map of part of Romania
	// from Page 68
	Arad->          add_edge_ptr( 140, Sibiu          );
	Arad->          add_edge_ptr( 118, Timisoara      );
	Arad->          add_edge_ptr(  75, Zerind         );
	Bucharest->     add_edge_ptr( 211, Fagaras        );
	Bucharest->     add_edge_ptr(  90, Giurgiu        );
	Bucharest->     add_edge_ptr( 101, Pitesti        );
	Bucharest->     add_edge_ptr(  85, Urziceni       );
	Craiova->       add_edge_ptr( 120, Drobeta        );
	Craiova->       add_edge_ptr( 138, Pitesti        );
	Craiova->       add_edge_ptr( 146, Rimnicu_Vilcea );
	Drobeta->       add_edge_ptr( 120, Craiova        );
	Drobeta->       add_edge_ptr(  75, Mehadia        );
	Eforie->        add_edge_ptr(  86, Hirsova        );
	Fagaras->       add_edge_ptr( 211, Bucharest      );
	Fagaras->       add_edge_ptr(  99, Sibiu          );
	Giurgiu->       add_edge_ptr(  90, Bucharest      );
	Hirsova->       add_edge_ptr(  86, Eforie         );
	Hirsova->       add_edge_ptr(  98, Urziceni       );
	Iasi->          add_edge_ptr(  87, Neamt          );
	Iasi->          add_edge_ptr(  92, Vaslui         );
	Lugoj->         add_edge_ptr(  70, Mehadia        );
	Lugoj->         add_edge_ptr( 111, Timisoara      );
	Mehadia->       add_edge_ptr(  75, Drobeta        );
	Mehadia->       add_edge_ptr(  70, Lugoj          );
	Neamt->         add_edge_ptr(  87, Iasi           );
	Oradea->        add_edge_ptr( 151, Sibiu          );
	Oradea->        add_edge_ptr(  71, Zerind         );
	Pitesti->       add_edge_ptr( 101, Bucharest      );
	Pitesti->       add_edge_ptr( 138, Craiova        );
	Pitesti->       add_edge_ptr(  97, Rimnicu_Vilcea );
	Rimnicu_Vilcea->add_edge_ptr( 146, Craiova        );
	Rimnicu_Vilcea->add_edge_ptr(  97, Pitesti        );
	Rimnicu_Vilcea->add_edge_ptr(  80, Sibiu          );
	Sibiu->         add_edge_ptr( 140, Arad           );
	Sibiu->         add_edge_ptr(  99, Fagaras        );
	Sibiu->         add_edge_ptr( 151, Oradea         );
	Sibiu->         add_edge_ptr(  80, Rimnicu_Vilcea );
	Timisoara->     add_edge_ptr( 118, Arad           );
	Timisoara->     add_edge_ptr( 111, Lugoj          );
	Urziceni->      add_edge_ptr(  85, Bucharest      );
	Urziceni->      add_edge_ptr(  98, Hirsova        );
	Urziceni->      add_edge_ptr( 142, Vaslui         );
	Vaslui->        add_edge_ptr(  92, Iasi           );
	Vaslui->        add_edge_ptr( 142, Urziceni       );
	Zerind->        add_edge_ptr(  75, Arad           );
	Zerind->        add_edge_ptr(  71, Oradea         );

	std::vector<std::shared_ptr<Vertex>> graph;
	graph.push_back( Arad           );
	graph.push_back( Bucharest      );
	graph.push_back( Craiova        );
	graph.push_back( Drobeta        );
	graph.push_back( Eforie         );
	graph.push_back( Fagaras        );
	graph.push_back( Giurgiu        );
	graph.push_back( Hirsova        );
	graph.push_back( Iasi           );
	graph.push_back( Lugoj          );
	graph.push_back( Mehadia        );
	graph.push_back( Neamt          );
	graph.push_back( Oradea         );
	graph.push_back( Pitesti        );
	graph.push_back( Rimnicu_Vilcea );
	graph.push_back( Sibiu          );
	graph.push_back( Timisoara      );
	graph.push_back( Urziceni       );
	graph.push_back( Vaslui         );
	graph.push_back( Zerind         );

	// Get source from command line; however, default to Arad.
	// The destination will always be Bucharest.
	std::shared_ptr<Vertex> source_vertex_ptr = Arad;
	std::string arg;
	for ( int i = 0; i < argc; ++i ) {
		arg = argv[ i ];
		if ( arg == std::string( "-source" ) ) {
			++i;
			arg = argv[ i ];
			auto graph_i = std::begin( graph );
			auto graph_end = std::end( graph );
			while ( graph_i != graph_end ) {
				auto vertex_ptr = *graph_i;
				if ( arg == vertex_ptr->get_name() ) {
					source_vertex_ptr = vertex_ptr;
					break;
				}
				++graph_i;
			}
		}
	}
	std::shared_ptr<Vertex> const destination_vertex_ptr = Bucharest;

	for_each_vertex_set_g_to_0( graph );
	for_each_vertex_set_h_to_heuristic( graph );

	//std::cout << "graph: " << graph << '\n';
	
	auto print_solution = [ &destination_vertex_ptr ]() {
		//solution using parents
		std::deque<std::shared_ptr<Vertex>> solution_ptrs;
		auto _ptr = destination_vertex_ptr;
		while ( _ptr != nullptr ) {
			solution_ptrs.push_front( _ptr );
			_ptr = _ptr->get_parent();
		}
		auto solution_ptrs_i = std::begin( solution_ptrs );
		auto solution_ptrs_end = std::end( solution_ptrs );
		if ( solution_ptrs_i != solution_ptrs_end ) {
			auto solution_ptr = *solution_ptrs_i;
			std::cout << "Solution:    " << solution_ptr->get_name();
			for ( ++solution_ptrs_i; solution_ptrs_i != solution_ptrs_end; ++solution_ptrs_i ) {
				auto solution_ptr = *solution_ptrs_i;
				std::cout << " -> " << solution_ptr->get_name();
			}
			std::cout << '\n';
		} 
	};
	(void)print_solution;
	
	auto is_equal_to = [] ( std::shared_ptr<Vertex> const & lhs_ptr, std::shared_ptr<Vertex> const & rhs_ptr ) -> bool {
		return lhs_ptr->get_name() == rhs_ptr->get_name();
	};
	(void)is_equal_to;

	auto by_lowest_f_first = []( std::shared_ptr<Vertex> const & lhs_ptr, std::shared_ptr<Vertex> const & rhs_ptr ) -> bool {
		return lhs_ptr->get_f() < rhs_ptr->get_f();
	};
	
	auto sort = [ &by_lowest_f_first ]( std::deque<std::shared_ptr<Vertex>> & sorted_ptrs ) {
		std::sort( std::begin( sorted_ptrs ), std::end( sorted_ptrs ), by_lowest_f_first );
	};
	
	auto push = [ &sort ]( std::deque<std::shared_ptr<Vertex>> & sorted_ptrs, std::shared_ptr<Vertex> const & ptr ) {
		sorted_ptrs.push_back( ptr );
		sort( sorted_ptrs );
	};
	(void)push;
	
	auto pop = [ &sort ]( std::deque<std::shared_ptr<Vertex>> & sorted_ptrs ) -> std::shared_ptr<Vertex> {
		auto ptr = sorted_ptrs.front();
		sorted_ptrs.pop_front();
		sort( sorted_ptrs );
		return ptr;
	};
	(void)pop;


	std::shared_ptr<Vertex> current_vertex_ptr;
	std::deque<std::shared_ptr<Vertex>> explored;
	std::deque<std::shared_ptr<Vertex>> frontier;
	
	auto is_in_explored = [ &explored ]( std::shared_ptr<Vertex> const & vertex_ptr ) -> bool {
		auto explored_end = std::end( explored );
		auto is_equal_to_vertex_ptr = [ &vertex_ptr ]( std::shared_ptr<Vertex> const & ptr ) -> bool {
			return ptr->get_name() == vertex_ptr->get_name();
		};
		return std::find_if( std::begin( explored ), explored_end, is_equal_to_vertex_ptr ) != explored_end;
	};
	(void)is_in_explored;
	
	auto is_in_frontier = [ &frontier ]( std::shared_ptr<Vertex> const & vertex_ptr ) -> bool {
		auto frontier_end = std::end( frontier );
		auto is_equal_to_vertex_ptr = [ &vertex_ptr ]( std::shared_ptr<Vertex> const & ptr ) -> bool {
			return ptr->get_name() == vertex_ptr->get_name();
		};
		return std::find_if( std::begin( frontier ), frontier_end, is_equal_to_vertex_ptr ) != frontier_end;
	};
	(void)is_in_frontier;

	auto find_in_frontier = [ &frontier ]( std::shared_ptr<Vertex> const & vertex_ptr ) -> std::deque<std::shared_ptr<Vertex>>::iterator {
		auto is_equal_to_vertex_ptr = [ &vertex_ptr ]( std::shared_ptr<Vertex> const & ptr ) -> bool {
			return ptr->get_name() == vertex_ptr->get_name();
		};
		return std::find_if( std::begin( frontier ), std::end( frontier ), is_equal_to_vertex_ptr );
	};
	(void)find_in_frontier;

	std::cout << "Source:      " << source_vertex_ptr->get_name() << '\n';
	std::cout << "Destination: " << destination_vertex_ptr->get_name() << '\n';
	std::cout << '\n';
	std::cout << "Expanded:\n";

	//-----------------------------------------------------------------------
	// Begin Your A* Algorithm Below:
	// Combine the Uniform-Cost-Search function on page 84 with the 
	// information on A* search starting on page 93.
	//-----------------------------------------------------------------------
	
	// ENTER YOUR A* SEARCH CODE HERE.
	
	//-----------------------------------------------------------------------
	// End Your A* Algorithm Above.
	//-----------------------------------------------------------------------

	//std::cout << "graph: " << graph << '\n';

	clear( graph );
	
}

